#!/usr/bin/python

# Copyright 2015 João Silva
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Development has been partly funded by the European Commission project
# QTLeap FP7-ICT-2013.4.1-610516 (http://qtleap.eu).

import sys
import re
from optparse import OptionParser
from nltk.tree import Tree
import itertools

leaf_regexp = re.compile(r'''(?x)
(?P<tok>.+)       # token
/(?P<pos>[^:|]+)   # pos
(?:
 (?:
  :(?P<lem>[^:|]+) # lemma (optional; always appears together with features)
 )?
 :(?P<feat>[^:|]+) # features (optional; might appear without lemma)
)?
:(?P<iob>[^:|]+)   # iob
\|(?P<head>\d+)
''')

# New format
leaf_regexp = re.compile(r'''
(?P<tok>\S+)              # token
(?:/(?P<lem>\S+))         # lemma (always present)
/(?P<pos>\S+?)            # pos
(?:\#(?P<feat>\S+))?      # inflection (optional)
\|(?P<head>\d+)
''', re.U|re.X)
#/(?P<bio>O|[IB]-\S+)      # bio (using slash separator)


class TreeDecorator:

    def __init__(self, tree, deps):
        self.tree = Tree(tree)
        self.deps = [line.split() for line in deps.split('\n')]

    @staticmethod
    def get_children(tree):
        return [tree[i] for i in range(len(tree))]

    def pick_head_using_deps(self, node):
        '''Heuristics for picking the head of a node (using dependencies).'''
        head = None
        deps = self.deps
        children = [c.node if isinstance(c, Tree) else c
                    for c in self.get_children(node)]
        children = [c.rsplit('|', 1)[1]
                    for c in children
                    if not c.startswith('PNT')]
        if len(children) == 2:
            a_cand, b_cand = children
            a_cand, b_cand = a_cand.split(','), b_cand.split(',')
            for a in a_cand:
                a = int(a)
                for b in b_cand:
                    b = int(b)
                    if deps[a][6] == deps[b][0]:
                        head = b
                        break
                    if deps[b][6] == deps[a][0]:
                        head = a
                        break
                    if deps[a][3] == 'PNT' and deps[b][3] != 'PNT':
                        head = b
                        break
                    if deps[b][3] == 'PNT' and deps[a][3] != 'PNT':
                        head = a
                        break
                    if deps[a][6] == '0' and deps[b][6] != '0':
                        head = b
                        break
                    if deps[b][6] == '0' and deps[a][6] != '0':
                        head = a
                        break
                if head is not None:
                    break
        else:
            head = ','.join(children)
        return head

    def pick_head_using_constituency(self, node):
        '''Heuristics for picking the head of a node (using constituency).'''
        head = None
        children = [c.node
                    for c in self.get_children(node)
                    if isinstance(c, Tree)]
        children = [c.rsplit('|', 1)
                    for c in children
                    if not c.startswith('PNT')]
        children = [h for (tag, h) in children
                    if tag.find('-') != -1]
        if len(children) == 1:
            head = int(children[0])
        return head

    def annotate_heads(self):
        '''Decorates nodes of the syntactic tree with the index of the head.'''
        leaf_counter = 0
        for t in self.tree.treepositions(order = 'postorder'):
            node = self.tree[t]
            if isinstance(node, Tree):
                head = self.pick_head_using_deps(node)
                if head is None:
                    head = self.pick_head_using_constituency(node)
                    if head is None:
                        raise Exception
                node.node += '|%s' % head
            else:
                parent = self.tree[t[:-1]]
                assert len(parent) == 1
                parent[0] = node + '|%s' % leaf_counter
                leaf_counter += 1


class FeatureExtractor:
    
    trees = None
    sentences = None
    markables = None
    
    def __init__(self):
        '''...'''
        self.trees = []
        self.tokens = []
        self.sentences = []
        self.markables = []
        
    def add(self, tree):
        '''...'''
        self.trees.append(tree)
        self.tokens.append([tok for tok in tree.leaves()])
        self.sentences.append([self.strip_token(tok) for tok in tree.leaves()])

    @staticmethod
    def get_phrasal_nodes(tree):
        '''...'''
        return [tree[t] for t in tree.treepositions(order = 'postorder')
                if isinstance(tree[t], Tree)]
    
    @staticmethod
    def strip_token(token):
        '''...'''
        return re.sub(leaf_regexp, '\g<tok>', token)
    
    @staticmethod
    def is_markable(n):
        '''...'''
        return n.node.startswith('NP') or n.node.startswith('PRS')

    def extract_markables(self):
        '''...'''
        markables = self.markables
        for n, tree in enumerate(self.trees):
            sentence_markables = [m for m in self.get_phrasal_nodes(tree)
                                  if self.is_markable(m)]
            markables.extend([(n, m) for m in sentence_markables])
        #print '%d markables' % len(markables)
        
    @staticmethod
    def head_to_slice(h):
        '''...'''
        temp = h.split(',')
        return slice(int(temp[0]), int(temp[-1]) + 1)

    def get_features(self):
        '''...'''
        for a, b in itertools.combinations(self.markables, 2):
            a_sent, a_tree = int(a[0]), a[1]
            b_sent, b_tree = int(b[0]), b[1]
            a_root = self.trees[a_sent]
            b_root = self.trees[b_sent]
            a_node, a_head = a_tree.node.rsplit('|', 1)
            b_node, b_head = b_tree.node.rsplit('|', 1)
            a_leaves = a_tree.leaves()
            b_leaves = b_tree.leaves()
            a_start = a_root.leaves().index(a_leaves[0])
            b_start = b_root.leaves().index(b_leaves[0])
            a_end = a_root.leaves().index(a_leaves[-1])
            b_end = b_root.leaves().index(b_leaves[-1])
            a_words = ' '.join([self.strip_token(tok) for tok in a_leaves])
            b_words = ' '.join([self.strip_token(tok) for tok in b_leaves])
            a_head_slice = self.head_to_slice(a_head)
            b_head_slice = self.head_to_slice(b_head)
            a_head_str = ' '.join(self.tokens[a_sent][a_head_slice])
            b_head_str = ' '.join(self.tokens[b_sent][b_head_slice])
            a_match = re.match(leaf_regexp, a_head_str)
            b_match = re.match(leaf_regexp, b_head_str)
            cores_match = self.strip_token(a_head_str) == self.strip_token(b_head_str)
            gender_agreement = a_match.group('feat') == b_match.group('feat')
            # print '\nantecedent: %s\nanaphora:   %s' % (a, b)
            # print '(%d) %s [%s - %s] <-%s-> (%d) %s [%s - %s]' \
            #     % (a_sent, a_words, a_head_str, a_match.group('feat'),
            #        cores_match and gender_agreement,
            #        b_sent, b_words, b_head_str, b_match.group('feat'))
            # print '%s %d:%d-%d %d:%d-%d' \
            #     % (cores_match and gender_agreement,
            #        a_sent, a_head_slice.start, a_head_slice.stop,
            #        b_sent, b_head_slice.start, b_head_slice.stop)
            print '%d:%d-%d %d:%d-%d %s' \
                % (a_sent, a_start, a_end,
                   b_sent, b_start, b_end,
                   cores_match and gender_agreement)


def treebank_read(filename):
    '''...'''
    with open(filename, 'r') as handle:
        for line in handle:
            yield line.strip()

def depbank_read(filename):
    '''...'''
    with open(filename, 'r') as handle:
        accumulator = []
        for line in handle:
            line = line.strip()
            if len(line) == 0:
                yield '\n'.join(accumulator)
                accumulator = []
            else:
                accumulator.append(line)
        if len(accumulator) > 0:
            yield '\n'.join(accumulator)

def handle_command_line():
    '''...'''
    usage = 'Usage: %prog [OPTIONS]'
    version = '%prog 0.1'
    description = '''This script is...'''
    parser = OptionParser(usage=usage, version=version, description=description)
    (options, args) = parser.parse_args()
    return (options, args)

def main():
    options, args = handle_command_line()
    extractor = FeatureExtractor()
    filename = args[0]
    treebank_filename = '%s.propbank' % filename
    depbank_filename = '%s.depbank' % filename
    try:
        treebank_reader = treebank_read(treebank_filename)
        depbank_reader = depbank_read(depbank_filename)
        sentence_count = 0
        while True:
            tree = treebank_reader.next()
            deps = depbank_reader.next()
            sentence_count += 1
            decorator = TreeDecorator(tree, deps)
            try:
                decorator.annotate_heads()
                extractor.add(decorator.tree)
            except Exception:
                print >> sys.stderr, '[%d] Failed to find head' % sentence_count
    except StopIteration:
        pass # To-Do: A better way to end
    extractor.extract_markables()
    extractor.get_features()

if __name__ == '__main__':
    main()
